@@ -160,7 +160,7 @@ void LatexParsingTest::test_latexLexing_data() {
                                   << (STypes() << T::none << T::text << T::text << T::none << T::none)
                                   << (Starts() << 0 << 7 << 8 << 13 << 14)
                                   << (Length() << 7 << 12 << 4 << 5 << 3)
-                                  << (Levels() << 0 << 1 << 1 << 1 << 1);
+                                  << (Levels() << 0 << 1 << 1 << 2 << 2);
     QTest::newRow("text command with nested square brackets") << "\\textbf{text [abc]}"
                                   << (TTypes() << T::command << T::braces << T::word << T::word)
                                   << (STypes() << T::none << T::text << T::text << T::text)
@@ -281,13 +281,13 @@ void LatexParsingTest::test_latexLexing_data() {
                                                   << (STypes() << T::none << T::keyValArg << T::none << T::keyVal_val << T::keyVal_val << T::imagefile << T::none)
                                                   << (Starts() << 0  << 16 << 17 << 21 << 22 << 28 << 29)
                                                   << (Length() << 16 << 12 << 3  <<  6 << 4  << 6  << 4)
-                                                  << (Levels() << 0  << 1  << 1  << 2  << 2  << 1  << 1);
+                                                  << (Levels() << 0  << 1  << 1  << 3  << 3  << 1  << 1);
     QTest::newRow("graphics command with keyval in braces, multiline") << "\\includegraphics[opt={\ntext\n}]{file}"
                                                   << (TTypes() << T::command << T::openSquare << T::keyVal_key << T::openBrace << T::word  << T::closeBrace << T::closeSquareBracket << T::braces << T::imagefile)
                                                   << (STypes() << T::none << T::keyValArg << T::none << T::keyVal_val << T::keyVal_val << T::keyVal_val << T::keyValArg << T::imagefile << T::none)
                                                   << (Starts() << 0  << 16 << 17 << 21 << 0 << 0 << 1 << 2 << 3 )
                                                   << (Length() << 16 << 6  << 3  <<  1 << 4 << 1 << 1 << 6 << 4 )
-                                                  << (Levels() << 0  << 1  << 1  << 2  << 3 << 3 << 2 << 1 << 1 );
+                                                  << (Levels() << 0  << 1  << 1  << 3  << 3 << 3 << 1 << 1 << 1 );
     QTest::newRow("listings command with defined keyval argument") << "\\lstdefinelanguage{Excel}{morekeywords={ab$c}}"
                                                                    << (TTypes() << T::command << T::braces     << T::word       << T::braces    << T::keyVal_key << T::braces )
                                                                    << (STypes() << T::none    << T::generalArg << T::generalArg << T::keyValArg << T::none       << T::definition)
@@ -306,7 +306,14 @@ void LatexParsingTest::test_latexLexing_data() {
         << (STypes() << T::none    << T::generalArg << T::generalArg << T::keyValArg << T::none       << T::definition<< T::definition << T::keyValArg  << T::none)
         << (Starts() << 0  << 18 << 19 << 25 << 26 << 39 << 0 << 1 << 3)
         << (Length() << 18 <<  7 <<  5 << 20 << 12 <<  6 << 1 << 1 << 4)
-        << (Levels() << 0  <<  1 <<  1 <<  1 <<  1 <<  3 << 3 << 2 << 0);
+        << (Levels() << 0  <<  1 <<  1 <<  1 <<  1 <<  3 << 3 << 1 << 0);
+    QTest::newRow("keyval argument multi line, no brace")
+        << "\\mycommand{note=\nabc\nsd\n} \n test"
+        << (TTypes() << T::command << T::openBrace << T::keyVal_key   << T::word       << T::word       << T::closeBrace << T::word)
+        << (STypes() << T::none    << T::keyValArg << T::none         << T::keyVal_val << T::keyVal_val << T::keyValArg  << T::none)
+        << (Starts() << 0  << 10 << 11 << 0 << 0 << 0 << 1)
+        << (Length() << 10 <<  6 <<  4 << 3 << 2 << 1 << 4)
+        << (Levels() << 0  <<  1 <<  1 << 2 << 2 << 1 << 0);
     QTest::newRow("include command") << "\\include{text dsf}"
                                      << (TTypes() << T::command << T::braces << T::file)
                                      << (STypes() << T::none << T::file << T::none)
@@ -361,6 +368,60 @@ void LatexParsingTest::test_latexLexing_data() {
                                                     << (Starts() << 0 << 8 << 9 << 14 << 15 <<  21)
                                                     << (Length() << 8 << 6 << 4 << 6 << 4 << 5)
                                                     << (Levels() << 0 << 1 << 1 << 1 << 1 << 0);
+    QTest::newRow("default overlay") << "\\againframe[<test>]"
+                                                   << (TTypes() << T::command << T::squareBracket << T::overlay)
+                                                   << (STypes() << T::none << T::overlay << T::none)
+                                                   << (Starts() << 0 << 11 << 12)
+                                                   << (Length() << 11 << 8 << 6)
+                                                   << (Levels() << 0 << 1 << 1 );
+    QTest::newRow("option") << "\\againframe[test]"
+                                     << (TTypes() << T::command << T::squareBracket << T::keyVal_key)
+                                     << (STypes() << T::none << T::keyValArg << T::none)
+                                     << (Starts() << 0 << 11 << 12)
+                                     << (Length() << 11 << 6 << 4)
+                                     << (Levels() << 0 << 1 << 1 );
+    QTest::newRow("keyval label") << "\\againframe[label=abc]"
+                                  << (TTypes() << T::command << T::squareBracket << T::keyVal_key<< T::label)
+                                  << (STypes() << T::none << T::keyValArg << T::none<< T::keyVal_val)
+                                  << (Starts() << 0 << 11 << 12 << 18)
+                                  << (Length() << 11 << 11 << 5 << 3)
+                                  << (Levels() << 0 << 1 << 1 << 2 );
+    QTest::newRow("keyval label in braces") << "\\againframe[label={abc}]"
+                                  << (TTypes() << T::command << T::squareBracket << T::keyVal_key<< T::braces << T::label)
+                                  << (STypes() << T::none << T::keyValArg << T::none<< T::label << T::none)
+                                  << (Starts() << 0 << 11 << 12 << 18 <<19)
+                                  << (Length() << 11 << 13 << 5 << 5 << 3 )
+                                  << (Levels() << 0 << 1 << 1 << 3 << 3);
+    QTest::newRow("keyval length") << "\\mycommand{width=\\textwidth}"
+                                  << (TTypes() << T::command << T::braces << T::keyVal_key<< T::command)
+                                  << (STypes() << T::none << T::keyValArg << T::none<< T::width)
+                                  << (Starts() << 0 << 10 << 11 << 17)
+                                  << (Length() << 10 << 18 << 5 << 10)
+                                  << (Levels() << 0 << 1 << 1 << 2 );
+    QTest::newRow("keyval text") << "\\mycommand{title=abc}"
+                                           << (TTypes() << T::command << T::braces << T::keyVal_key<< T::word)
+                                           << (STypes() << T::none << T::keyValArg << T::none<< T::text)
+                                           << (Starts() << 0 << 10 << 11 << 17)
+                                           << (Length() << 10 << 11 << 5 << 3)
+                                           << (Levels() << 0 << 1 << 1 << 2);
+    QTest::newRow("keyval text, multiple words, no brace") << "\\mycommand{title=abc def}"
+                                 << (TTypes() << T::command << T::braces << T::keyVal_key<< T::word<< T::word)
+                                 << (STypes() << T::none << T::keyValArg << T::none<< T::text<< T::text)
+                                 << (Starts() << 0 << 10 << 11 << 17<< 21)
+                                 << (Length() << 10 << 15 << 5 << 3 << 3)
+                                 << (Levels() << 0 << 1 << 1 << 2 << 2);
+    QTest::newRow("keyval text in braces") << "\\mycommand{title={abc}}"
+                                   << (TTypes() << T::command << T::braces << T::keyVal_key<< T::braces << T::word)
+                                   << (STypes() << T::none << T::keyValArg << T::none<< T::text << T::text)
+                                   << (Starts() << 0 << 10 << 11 << 17 << 18)
+                                   << (Length() << 10 << 13 << 5 << 5 << 3)
+                                   << (Levels() << 0 << 1 << 1 << 3 << 3);
+    QTest::newRow("keyval text followed by second key/val") << "\\mycommand{title=abc,note={abc}}"
+                                                           << (TTypes() << T::command << T::braces << T::keyVal_key<< T::word<< T::keyVal_key<< T::braces << T::word)
+                                                           << (STypes() << T::none << T::keyValArg << T::none<< T::text<< T::none << T::keyVal_val << T::keyVal_val)
+                                                           << (Starts() << 0 << 10 << 11 << 17<< 21<<26<<27)
+                                                           << (Length() << 10 << 22 << 5 << 3 << 4<<5 <<3)
+                                                           << (Levels() << 0 << 1 << 1 << 2  <<1 << 3 << 3);
 
 }
 
@@ -371,7 +432,7 @@ void LatexParsingTest::test_latexLexing() {
     lp->commandDefs.unite(pkg_graphics.commandDescriptions);
     LatexPackage pkg_tex = loadCwlFile("tex.cwl");
     lp->commandDefs.unite(pkg_tex.commandDescriptions);
-    LatexPackage pkg_listings = loadCwlFile("listings.cwl");
+    LatexPackage pkg_listings = loadCwlFile("txs-test.cwl");
     lp->commandDefs.unite(pkg_listings.commandDescriptions);
     LatexPackage pkg_yagu = loadCwlFile("yagusylo.cwl");
     lp->commandDefs.unite(pkg_yagu.commandDescriptions);
@@ -554,7 +615,6 @@ void LatexParsingTest::test_getArg_data() {
                                             <<   (STypes() <<T::package) << (QStringList() <<"siunitx,test2, test3");
     QTest::newRow("usepackage command, multilines with comment") << "\\usepackage{siunitx,\ntest2,%abc\ntest3}"
                                             <<   (STypes() <<T::package) << (QStringList() <<"siunitx,test2,test3");
-
 }
 
 void LatexParsingTest::test_getArg() {
@@ -702,6 +762,11 @@ void LatexParsingTest::test_getArg2_data() {
                                             << (ATypes() <<ArgumentList::Mandatory)
                                             << (QList<int>()<<0)
                                             << (QStringList() <<"hyperref");
+    // issue 2843
+    QTest::newRow("text command, multi-line with keyval") << "\\documentclass[xcolor={table,\nsvgnames}]{beamer}"
+                                             << (ATypes() <<ArgumentList::Mandatory)
+                                             << (QList<int>()<<0)
+                                             << (QStringList() <<"beamer");
 
 }
 
@@ -744,6 +809,78 @@ void LatexParsingTest::test_getArg2() {
     ConfigManager::RUNAWAYLIMIT=storeRUNAWAYLIMIT;
 }
 
+void LatexParsingTest::test_getArgContent_data()
+{
+    QTest::addColumn<QString>("lines");
+    QTest::addColumn<STypes>("types");
+    QTest::addColumn<STypes>("subtypes");
+    QTest::addColumn<int>("pos");
+    QTest::addColumn<int>("level");
+
+    QTest::newRow("newcommand command") << "\\newcommand{text}{test}"
+                                        << (TTypes() <<T::braces << T::def << T::braces)
+                                        << (STypes() <<T::def << T::none << T::definition)
+                                        << 0 << 1;
+    QTest::newRow("newcommand command2") << "\\newcommand{\\ext}{test}"
+                                         <<  (STypes() <<T::braces << T::def << T::braces)
+                                         <<  (STypes() <<T::def << T::none << T::definition)
+                                         << 0 << 1;
+    QTest::newRow("newcommand command, no braces") << "\\newcommand text {test}"
+                                                   <<  (STypes() <<T::word << T::braces)
+                                                   <<  (STypes() <<T::def << T::definition)
+                                                   << 0 << 1;
+    QTest::newRow("newcommand command, no braces2") << "\\newcommand text test"
+                                                    <<  (TTypes() << T::word << T::word)
+                                                    <<  (STypes() << T::def << T::definition)
+                                                    << 0 << 1;
+    QTest::newRow("newcommand command with trail") << "\\newcommand{text}{test} abc"
+                                        << (TTypes() <<T::braces << T::def << T::braces)
+                                        << (STypes() <<T::def << T::none << T::definition)
+                                        << 0 << 1;
+    QTest::newRow("newcommand command with trailing braces") << "\\newcommand{text}{test} {abc}"
+                                                   << (TTypes() <<T::braces << T::def << T::braces)
+                                                   << (STypes() <<T::def << T::none << T::definition)
+                                                   << 0 << 1;
+}
+
+void LatexParsingTest::test_getArgContent()
+{
+    QSharedPointer<LatexParser> lp = QSharedPointer<LatexParser>::create();
+    *lp=LatexParser::getInstance();
+    LatexPackage pkg_graphics = loadCwlFile("graphicx.cwl");
+    lp->commandDefs.unite(pkg_graphics.commandDescriptions);
+    QFETCH(QString,lines);
+    QFETCH(TTypes, types);
+    QFETCH(STypes, subtypes);
+    QFETCH(int, pos);
+    QFETCH(int, level);
+
+    QDocument *doc = new QDocument();
+    doc->setText(lines, false);
+    for(int i=0; i<doc->lines(); i++){
+        QDocumentLineHandle *dlh = doc->line(i).handle();
+        Parsing::simpleLexLatexLine(dlh);
+    }
+    TokenStack stack;
+    CommandStack commandStack;
+    for(int i=0; i<doc->lines(); i++){
+        QDocumentLineHandle *dlh = doc->line(i).handle();
+        Parsing::latexDetermineContexts2(dlh, stack, commandStack, lp);
+    }
+    QDocumentLineHandle *dlh = doc->line(0).handle();
+    TokenList tl= dlh->getCookieLocked(QDocumentLine::LEXER_COOKIE).value<TokenList >();
+    // first token is command
+    Token tkCmd;
+    TokenList args;
+    TokenList result=Parsing::getArgContent(tl, pos,level);
+    QCOMPARE(result.length(), types.length());
+    for(int i=0;i<result.length();++i){
+        QCOMPARE(result.at(i).type, types.at(i));
+        QCOMPARE(result.at(i).subtype, subtypes.at(i));
+    }
+    delete doc;
+}
+
 void LatexParsingTest::test_getTokenAtCol_data() {
     QTest::addColumn<QString>("lines");
     QTest::addColumn<QList<int> >("nr");
@@ -820,7 +957,7 @@ void LatexParsingTest::test_getCommandFromToken_data() {
     QTest::addColumn<QString>("desiredResult");
 
 
-    QTest::newRow("simple") << "bummerang  \\test"
+    /*QTest::newRow("simple") << "bummerang  \\test"
                             << 0 << 0
                             << "";
 
@@ -843,10 +980,19 @@ void LatexParsingTest::test_getCommandFromToken_data() {
                             << "";
     QTest::newRow("simple7") << "bummerang  \\section{abc {cde}}"
                             << 3 << 0
-                            << "\\section";
-    QTest::newRow("simple8") << "bummerang  \\section{abc {cde}}"
+                            << "\\section";*/
+    QTest::newRow("non-arg braces") << "bummerang  \\section{abc {cde}}"
                             << 4 << 0
                             << "\\section";
+    QTest::newRow("non-arg braces 2") << "bummerang  \\section{abc {cde} fgh}"
+                             << 6 << 0
+                             << "\\section";
+    QTest::newRow("non-arg braces 3") << "bummerang  \\section{abc {cde fgh}}"
+                                    << 6 << 0
+                                    << "\\section";
+    QTest::newRow("non-arg braces, nested") << "bummerang  \\section{abc {cde {fgh}}}"
+                                    << 6 << 0
+                                    << "\\section";
     QTest::newRow("optonal") << "bummerang  \\section[ab ab]{abc cde}"
                             << 2 << 0
                             << "\\section";
@@ -898,6 +1044,9 @@ void LatexParsingTest::test_getCommandFromToken_data() {
     QTest::newRow("multi-line") << "bummerang  \\section{abc\n cde}"
                             << 0 << 1
                             << "\\section";
+    QTest::newRow("multi-line, brace in new line") << "bummerang  \\section\n{abc cde}"
+                                << 0 << 1
+                                << "\\section";
 
 
 }
@@ -944,10 +1093,30 @@ void LatexParsingTest::test_getContext_data() {
                             << 2
                             << (TTypes() << T::word)
                             << (STypes() << T::none);
+    QTest::newRow("none") << "abc   abc"
+                            << 4
+                            << (TTypes() )
+                            << (STypes());
     QTest::newRow("command") << "\\section{abc}"
                             << 10
                             << (TTypes() << T::command <<T::braces<<T::word)
                             << (STypes() << T::none << T::title<<T::title);
+    QTest::newRow("command, unclosed brace") << "\\section{abc"
+                             << 10
+                             << (TTypes() << T::command <<T::openBrace<<T::word)
+                             << (STypes() << T::none << T::title<<T::title);
+    QTest::newRow("command, unclosed brace with spaces") << "\\section{abc   "
+                             << 13
+                             << (TTypes() << T::command <<T::openBrace<<T::word)
+                             << (STypes() << T::none << T::title<< T::title);
+    QTest::newRow("after command") << "\\section{abc}  "
+                             << 14
+                             << (TTypes())
+                             << (STypes());
+    QTest::newRow("after command on word") << "\\section{abc} abc"
+                           << 16
+                           << (TTypes() << T::word)
+                           << (STypes() << T::none);
     QTest::newRow("command without braces") << "\\section abc"
                             << 10
                             << (TTypes() << T::command << T::word)
@@ -956,14 +1125,20 @@ void LatexParsingTest::test_getContext_data() {
                             << 10
                             << (TTypes() << T::command << T::squareBracket<<T::word)
                             << (STypes() << T::none << T::shorttitle<<T::shorttitle);
+    QTest::newRow("command with optional arg 2") << "\\section[fds]{abc}"
+                             << 15
+                             << (TTypes() << T::command <<T::braces<<T::word)
+                             << (STypes() << T::none << T::title<<T::title);
     QTest::newRow("command with keyval") << "\\includegraphics[width=4cm]{abc}"
                             << 18
                             << (TTypes() << T::command << T::squareBracket<<T::keyVal_key)
                             << (STypes() << T::none << T::keyValArg<<T::none);
     QTest::newRow("command with keyval2") << "\\includegraphics[width=4cm]{abc}"
                             << 23
-                            << (TTypes() << T::command << T::squareBracket<<T::keyVal_key<<T::width)
-                            << (STypes() << T::none << T::keyValArg<<T::none<<T::keyVal_val);
+                            << (TTypes() << T::command << T::squareBracket << T::keyVal_key)
+                            << (STypes() << T::none    << T::keyValArg     << T::none);
+                            //<< (TTypes() << T::command << T::squareBracket<<T::keyVal_key<<T::width) // this may be the desired outcome
+                            //<< (STypes() << T::none << T::keyValArg<<T::none<<T::keyVal_val);
     QTest::newRow("command with keyval3") << "\\includegraphics[width=4cm]{abc}"
                             << 24
                             << (TTypes() << T::command << T::squareBracket<<T::keyVal_key<<T::width)
@@ -972,17 +1147,36 @@ void LatexParsingTest::test_getContext_data() {
                             << 28
                             << (TTypes() << T::command << T::squareBracket<<T::keyVal_key)
                             << (STypes() << T::none << T::keyValArg<<T::none);
+    QTest::newRow("command with keyval5") << "\\includegraphics[width as=4cm]{abc}"
+                          << 24
+                          << (TTypes() << T::command << T::squareBracket<<T::keyVal_key)
+                          << (STypes() << T::none << T::keyValArg<<T::none);
+    QTest::newRow("command with empty keyval") << "\\includegraphics[width=]{abc}"  // #4017
+                            << 23
+                            << (TTypes() << T::command << T::squareBracket<<T::keyVal_key<<T::keyVal_val)
+                            << (STypes() << T::none << T::keyValArg<<T::none<<T::keyVal_val);
+    QTest::newRow("command with empty keyval and space") << "\\includegraphics[width= ]{abc}"  // #4218
+                           << 23
+                           << (TTypes() << T::command << T::squareBracket<<T::keyVal_key)
+                           << (STypes() << T::none << T::keyValArg<<T::none);
     QTest::newRow("command with keyval as defined argument") << "\\lstdefinelanguage{Excel}{morekeywords={ab$c}}"
-                                          << 41
-                                                             << (TTypes() << T::command << T::braces<<T::keyVal_key<<T::braces)
-                                                             << (STypes() << T::none << T::keyValArg<<T::none<<T::definition);
+                            << 41
+                             << (TTypes() << T::command << T::braces<<T::keyVal_key<<T::braces)
+                             << (STypes() << T::none << T::keyValArg<<T::none<<T::definition);
     QTest::newRow("command with keyval as label") << "\\lstdefinelanguage{Excel}{label=test}"
-                                                             << 35
-                                                             << (TTypes() << T::command << T::braces<<T::keyVal_key<<T::label)
-                                                             << (STypes() << T::none << T::keyValArg<<T::none<<T::keyVal_val);
-
-
-
+                             << 35
+                             << (TTypes() << T::command << T::braces<<T::keyVal_key<<T::label)
+                             << (STypes() << T::none << T::keyValArg<<T::none<<T::keyVal_val);
+    QTest::newRow("following command") << "bummerang\\text" // command after word, #3967
+                            << 9
+                            << (TTypes() << T::word)
+                            << (STypes() << T::none);
+    // stacked commands, keyval with arguments which are comma separated (#4074)
+    QTest::newRow("command with keyval with command with comma separated arguments")
+        << "\\mycommand{note=\\cite{abc}}"
+        << 23
+        << (TTypes() << T::command << T::braces<<T::keyVal_key<<T::command<<T::braces<<T::bibItem)
+        << (STypes() << T::none << T::keyValArg<<T::none<<T::keyVal_val<<T::bibItem<<T::none);
 }
 
 void LatexParsingTest::test_getContext() {
@@ -990,7 +1184,7 @@ void LatexParsingTest::test_getContext() {
     *lp=LatexParser::getInstance();
     LatexPackage pkg_graphics = loadCwlFile("graphicx.cwl");
     lp->commandDefs.unite(pkg_graphics.commandDescriptions);
-    LatexPackage pkg_listings = loadCwlFile("listings.cwl");
+    LatexPackage pkg_listings = loadCwlFile("txs-test.cwl");
     lp->commandDefs.unite(pkg_listings.commandDescriptions);
     QFETCH(QString,lines);
     QFETCH(int, nr);
@@ -1015,9 +1209,116 @@ void LatexParsingTest::test_getContext() {
     TokenStack result = Parsing::getContext(dlh,nr);
 
     for(int k=0;k<result.size();k++){
+        if(k>=desiredResults.size()){
+            continue;
+        }
         QVERIFY2(result.at(k).type==desiredResults.at(k), QString("incorrect type at index %1:%2").arg(k).arg(lines).toLatin1());
         QVERIFY2(result.at(k).subtype==types.at(k), QString("incorrect subtype at index %1:%2").arg(k).arg(lines).toLatin1());
     }
+    QVERIFY2(result.size()==desiredResults.size(), QString("incorrect stacksize: %1:%2").arg(result.size()).arg(desiredResults.size()).toLatin1());
+
+    delete doc;
+}
+
+void LatexParsingTest::test_getContextMultiLine_data() {
+    QTest::addColumn<QString>("lines");
+    QTest::addColumn<int >("nr");
+    QTest::addColumn<int >("ln");
+    QTest::addColumn<TTypes>("desiredResults");
+    QTest::addColumn<STypes>("types");
+
+    QTest::newRow("simple") << "bummerang\nbummerang"
+                            << 2 << 1
+                            << (TTypes() << T::word)
+                            << (STypes() << T::none);
+    QTest::newRow("command") << "\\section{\nabc}"
+                             << 1 << 1
+                             << (TTypes() <<T::openBrace<<T::word)
+                             << (STypes() << T::title<<T::title);
+    QTest::newRow("command, unclosed brace") << "\\section{abc\nabc\n}"
+                             << 2 << 1
+                             << (TTypes() << T::openBrace<<T::word)
+                             << (STypes() << T::title<<T::title);
+    QTest::newRow("command, unclosed brace with spaces") << "\\section{abc\n   "
+                             << 1 << 1
+                             << (TTypes() << T::openBrace)
+                             << (STypes() << T::title);
+    QTest::newRow("after command") << "\\section{abc}\n  "
+                           << 1 << 1
+                           << (TTypes())
+                           << (STypes());
+    QTest::newRow("after command on word") << "\\section{abc}\n abc"
+                               << 2 << 1
+                               << (TTypes() << T::word)
+                               << (STypes() << T::none);
+    QTest::newRow("command with keyval") << "\\includegraphics[\nwidth=4cm]{abc}"
+                             << 1 << 1
+                             << (TTypes() << T::openSquare<<T::keyVal_key)
+                             << (STypes() << T::keyValArg<<T::none);
+    QTest::newRow("command with keyval2") << "\\includegraphics[\nwidth=4cm]{abc}"
+                              << 7 << 1
+                              << (TTypes() << T::openSquare << T::keyVal_key << T::width)
+                              << (STypes() << T::keyValArg     << T::none << T::keyVal_val);
+    QTest::newRow("command with keyval3") << "\\includegraphics\n[\nwidth=4cm]{abc}"
+                                         << 1 << 2
+                                         << (TTypes() << T::openSquare<<T::keyVal_key)
+                                         << (STypes() << T::keyValArg<<T::none);
+    QTest::newRow("command with keyval4") << "\\includegraphics\n[\nwidth=4cm]{abc}"
+                                  << 7 << 2
+                                  << (TTypes() << T::openSquare << T::keyVal_key << T::width)
+                                  << (STypes() << T::keyValArg     << T::none << T::keyVal_val);
+    QTest::newRow("command with keyval5") << "\\includegraphics\n[\nwidth=      ]{abc}"
+                                          << 7 << 2
+                                          << (TTypes() << T::openSquare<<T::keyVal_key)
+                                          << (STypes() << T::keyValArg<<T::none);
+    QTest::newRow("after command, open brace") << "\\section{\nabc}  "
+                                   << 5 << 1
+                                   << (TTypes())
+                                   << (STypes());
+}
+
+void LatexParsingTest::test_getContextMultiLine() {
+    QSharedPointer<LatexParser> lp = QSharedPointer<LatexParser>::create();
+    *lp=LatexParser::getInstance();
+    LatexPackage pkg_graphics = loadCwlFile("graphicx.cwl");
+    lp->commandDefs.unite(pkg_graphics.commandDescriptions);
+    LatexPackage pkg_listings = loadCwlFile("txs-test.cwl");
+    lp->commandDefs.unite(pkg_listings.commandDescriptions);
+    QFETCH(QString,lines);
+    QFETCH(int, nr);
+    QFETCH(int, ln);
+    QFETCH(TTypes, desiredResults);
+    QFETCH(STypes, types);
+
+    QDocument *doc = new QDocument();
+    doc->setText(lines, false);
+    for(int i=0; i<doc->lines(); i++){
+        QDocumentLineHandle *dlh = doc->line(i).handle();
+        Parsing::simpleLexLatexLine(dlh);
+    }
+    TokenStack stack;
+    CommandStack commandStack;
+    for(int i=0; i<doc->lines(); i++){
+        QDocumentLineHandle *dlh = doc->line(i).handle();
+        Parsing::latexDetermineContexts2(dlh, stack, commandStack, lp);
+    }
+    if(doc->lineCount()>ln){
+        QDocumentLineHandle *dlh = doc->line(ln).handle();
+        //TokenList tl= dlh->getCookieLocked(QDocumentLine::LEXER_COOKIE).value<TokenList >();
+
+        TokenStack result = Parsing::getContext(dlh,nr);
+
+        for(int k=0;k<result.size();k++){
+            if(k>=desiredResults.size()){
+                continue;
+            }
+            QVERIFY2(result.at(k).type==desiredResults.at(k), QString("incorrect type at index %1:%2").arg(k).arg(lines).toLatin1());
+            QVERIFY2(result.at(k).subtype==types.at(k), QString("incorrect subtype at index %1:%2").arg(k).arg(lines).toLatin1());
+        }
+        QVERIFY2(result.size()==desiredResults.size(), QString("incorrect stacksize: %1:%2").arg(result.size()).arg(desiredResults.size()).toLatin1());
+    }else{
+        qDebug()<<"test case lines !";
+    }
 
     delete doc;
 }
